/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2007-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi.opt.oo;

import java.io.IOException;

import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.type.BasicInterfaceType;
import org.objectweb.fractal.juliac.ClassGeneratorItf;
import org.objectweb.fractal.juliac.ITBasedClassGenerator;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.JuliacLoader;
import org.objectweb.fractal.juliac.conf.JuliacConfig;
import org.objectweb.fractal.juliac.desc.MembraneDesc;
import org.objectweb.fractal.juliac.proxy.InterceptorSourceCodeGenerator;
import org.objectweb.fractal.juliac.proxy.InterfaceImplementationClassGenerator;
import org.objectweb.fractal.juliac.proxy.ProxyClassGenerator;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;
import org.osoa.sca.annotations.Callback;
import org.ow2.frascati.tinfi.TinfiDomain;

/**
 * This class generates the source code associated to Fractal components.
 * The membrane implementation and the initializer implementation are generated.
 * 
 * The content, the interceptors and the controllers are kept in separate
 * classes and the controllers are implemented with objects.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 */
public class FCOOCtrlSourceCodeGenerator
extends org.objectweb.fractal.juliac.opt.oo.FCOOCtrlSourceCodeGenerator {
    
    public FCOOCtrlSourceCodeGenerator() {
        super();
    }

    /** Configuration file for Tinfi. */
    public static final String DEFAULT_CONFIGURATION = "tinfi-bundled.cfg";

    
    // -----------------------------------------------------------------------
    // Implementation of the FCSourceCodeGeneratorItf interface
    // -----------------------------------------------------------------------

    /**
     * Initialize this generator.
     * 
     * The initialization process is kept separated from the constructor in
     * order to be able to customize the julia.config system property.
     * 
     * @param jc  the instance of Juliac using this source code generator
     */
    @Override
    public void init( Juliac jc ) throws IOException {

        JuliacLoader loader = jc.getJuliacConfig().getJuliacLoader();
        loader.loadConfigFile(DEFAULT_CONFIGURATION);
        
        super.init(jc);

        /*
         * Register component interceptor code generators.
         */
        JuliacConfig jconf = jc.getJuliacConfig();
        jconf.putInterceptorSourceCodeGenerator(
            "org.ow2.frascati.tinfi.asm.SCAContentInterceptorCodeGenerator",
            new SCAContentInterceptorSourceCodeGenerator() );
        registerIntentInterceptorSourceCodeGenerator(jconf);
    }
    
    /**
     * Register the intent interceptor source code generators associated with
     * this optimization level source code generator. This method is kept
     * separated from {@link #init(Juliac)} in order to allow overriding in
     * {@link FCOOCtrlNoIntentSourceCodeGenerator}.
     * 
     * @since 1.1.1
     */
    protected void registerIntentInterceptorSourceCodeGenerator(
        JuliacConfig jconf ) {
        
        jconf.putInterceptorSourceCodeGenerator(
            "org.ow2.frascati.tinfi.asm.SCAIntentInterceptorCodeGenerator",
            new SCAIntentInterceptorSourceCodeGenerator() );
    }

    /**
     * Return <code>true</code> if the current source code generator handles the
     * specified controller descriptor.
     * 
     * @param controllerDesc  the controller descriptor to be checked
     * @return                if <code>controllerDesc</code> should be accepted 
     * @since 1.0
     */
    @Override
    public boolean acceptCtrlDesc( Object controllerDesc ) {
        boolean accept =
            controllerDesc.equals(TinfiDomain.SCA_PRIMITIVE) ||
            controllerDesc.equals(TinfiDomain.SCA_COMPOSITE);
        return accept;
    }
    
    /**
     * Return the source code generator for the initializer class associated
     * with this component source code generator.
     */
    @Override
    public InitializerOOCtrlClassGenerator getInitializerClassGenerator() {
        return new InitializerOOCtrlClassGenerator();
    }

    /**
     * Return the source code generator for component interfaces.
     */
    @Override
    public ProxyClassGenerator getInterfaceClassGenerator( 
        InterfaceType it, MembraneDesc<?> membraneDesc ) {

        ProxyClassGenerator pcg = null;
        String itname = it.getFcItfName();
        
        /*
         * See the comments in method generateInterfaceImpl for the rationale
         * for choosing interface source code generator.
         */
        
        if( itname.endsWith("-controller") || itname.equals("component") ) {
            pcg = new InterfaceImplementationClassGenerator(it,membraneDesc,false,jc);
        }
        else {
            if( it.isFcClientItf() ) {
                /*
                 * In theory we should be returning here two interface source
                 * code generator since the client interface class extends the
                 * server one corresponding to the same type. However, the
                 * current method is mainly used for retrieving the name of the
                 * generated interface class (such as in initializer source code
                 * generators.) It is then sufficient to return only the class
                 * which corresponds to the real implementation.
                 */
                pcg = new ClientInterfaceClassGenerator(it,membraneDesc,false,jc);
            }
            else {
                pcg = new ServerInterfaceClassGenerator(it,membraneDesc,false,jc);                
            }
        }
        
        return pcg;
    }
    
    /**
     * Generate the source code of component interfaces associated with the
     * specified interface type.
     */
    @Override
    protected void generateInterfaceImpl( InterfaceType it, String ctrlDesc )
    throws IOException {
        
        String itname = it.getFcItfName();
        if( itname.endsWith("-controller") || itname.equals("component") ) {

            /*
             * Control interfaces.
             * 
             * It seems to be wiser to keep
             * InterfaceImplementationClassGenerator as a source code generator
             * for control interfaces even though ServerInterfaceClassGenerator
             * could be used. This allows reusing pre-generated implementations
             * for Julia control interfaces.
             */
            ClassGeneratorItf cg = 
                new InterfaceImplementationClassGenerator(it,null,false,jc);
            jc.generateSourceCode(cg);
        }
        else {
           
            /*
             * Business interfaces.
             */

            // Server (input) interfaces
            ClassGeneratorItf cg =
                new ServerInterfaceClassGenerator(it,null,false,jc);
            jc.generateSourceCode(cg);
            
            if( it.isFcClientItf() ) {
                /*
                 * Client (output) interfaces.
                 * Server (input) interfaces are generated also for client ones
                 * since client interface classes extend the server class
                 * corresponding to the same type.
                 */
                cg = new ClientInterfaceClassGenerator(it,null,false,jc);
                jc.generateSourceCode(cg);
            }
                
            // ServiceReference implementation
            cg = new ServiceReferenceClassGenerator(it,null,false,jc);
            jc.generateSourceCode(cg);
            
            // CallableReference and input interface implementations for callbacks
            String signature = it.getFcItfSignature();
            UnifiedClass cl = jc.create(signature);
            Callback annot = cl.getAnnotation(Callback.class);
            if( annot != null ) {
                Class<?> cbcl = annot.value();
                InterfaceType cbit =
                    new BasicInterfaceType(
                        "callback", cbcl.getName(), false, false, false );
                cg = new CallableReferenceClassGenerator(cbit,null,false,jc);
                jc.generateSourceCode(cg);
                
                cg = new CallBackInterfaceClassGenerator(cbit,null,false,jc);
                jc.generateSourceCode(cg);
            }
        }
    }

    /**
     * Return the class generator for interceptors.
     */
    @Override
    protected ITBasedClassGenerator getInterceptorClassGenerator(
        InterceptorSourceCodeGenerator[] iscgs ) {
        return new InterceptorClassGenerator(iscgs,jc);
    }    
}
